#include "TTree.h"
#include "TFile.h"
#include "TH1.h"
#include "TH2.h"
#include "TF1.h"
#include "TRandom3.h"
using namespace std;
#include "treeMaker.h"

struct node_t{
  //xVec contents are pointers to the activations 
  //in the previous layer (layer l-1)
  vector<double*> xVec;
  vector<double> wVec;
  double y;
  double z;
  double a;
  double b;
  double delta;
};
struct layer_t{
  vector<node_t> nodeVec;
};

void printUsage(); // print usage   
double sigmoid(double z);
double sigmoidPrime(double z);

int main(int argc, char **argv){
  
  //char  hId[80];
  char  inFileName[150];
  char  outFileName[150];
  char *argptr;
  
  //Set the default output file                                                 
  sprintf(outFileName,"./outFile.root");
  if (argc == 1) printUsage();
  for (int i=1; i<argc; i++) {
    argptr = argv[i];
    if (*argptr == '-') {
      argptr++;
      switch (*argptr) {
      case 'h':
        printUsage();
        break;
      case 'o':
        strcpy(outFileName,++argptr);
        break;
      default:
        fprintf(stderr,"\nUnrecognized argument: [-%s]\n",argptr);
        printUsage();
        break;
      }
    } else {
      strcpy(inFileName,argptr);
    }
  }

  TH2D *hOut0 = new TH2D("hOut0","",1000,0.0,1.0,1000,0.0,1.0);
  TH2D *hOut1 = new TH2D("hOut1","",1000,0.0,1.0,1000,0.0,1.0);
  TH2D *hOut2 = new TH2D("hOut2","",1000,0.0,1.0,1000,0.0,1.0);
  TH2D *hOut3 = new TH2D("hOut3","",2000,0.0,1.0,1000,0.0,1.0);

  //Setup the random generator
  UInt_t mySeed = 3;
  TRandom3 *myRandGen = new TRandom3(mySeed);

  //Setup the number of nodes in each neural net layer 
  int n0=2;//Number of nodes in the input layer (l=0)
  int n1=2;//Number of nodes in the hidden layer (l=1)
  int n2=2;//Number of nodes in the ouput layer (l=2)

  //Create a vector that holds the number of nodes per layer
  vector<int> netDim;
  netDim.push_back(n0);
  netDim.push_back(n1);
  netDim.push_back(n2);

  vector<layer_t>layerVec;//This vector contains all the layers for the net
  layer_t layer;//This is just a template of a single layer that contains a vector for nodes
  node_t node;//This is just a template of a single node

  //So we have a layerVec that contains layers, where each layer contains nodes
  //and each node contains the remaining information 

  int nLayers = netDim.size();
  for (int layerIndex = 0; layerIndex < nLayers; layerIndex++) {
    layerVec.push_back(layer);//Create the layer
    for (int nodeIndex = 0; nodeIndex < netDim[layerIndex]; nodeIndex++) {
      layerVec[layerIndex].nodeVec.push_back(node);//Create the nodes within each layer
      //Do some initialization of the node
      layerVec[layerIndex].nodeVec[nodeIndex].a = 0.0;
      layerVec[layerIndex].nodeVec[nodeIndex].b = myRandGen->Uniform(0.0,1.0);//Important
      //to define each node differently. Otherwise, you can get stuck in a poor solution
      if (layerIndex > 0) {//For layers above the input layer
	//For each node (in l-1 = M1) feeding into the current layer. Note M1 means minus one
	for (int nodeIndexM1 = 0; nodeIndexM1 < netDim[layerIndex-1]; nodeIndexM1++){
	  //Create the weights
	  layerVec[layerIndex].nodeVec[nodeIndex].wVec.push_back(myRandGen->Uniform(0.0,1.0));
	  double *aPtr = &layerVec[layerIndex-1].nodeVec[nodeIndexM1].a;
	  layerVec[layerIndex].nodeVec[nodeIndex].xVec.push_back(aPtr);//Create the inputs
	}
      } 
    }
  }
  int outputLayerIndex = nLayers - 1;  
  cout<<"layerVec.size() = "<<layerVec.size()<<endl;
  cout<<"layerVec[0].nodeVec.size() = "<<layerVec[0].nodeVec.size()<<endl;
  cout<<"layerVec[1].nodeVec.size() = "<<layerVec[1].nodeVec.size()<<endl;
  cout<<"layerVec[2].nodeVec.size() = "<<layerVec[2].nodeVec.size()<<endl;

  //Setup the input tree
  TFile* inFile=new TFile(inFileName);
  TTree *inTree = (TTree*)inFile->Get("myTree");
  myTree_t myInputTree;
  getBranches(inTree,&myInputTree);

  //Setup the output file
  TFile outFile(outFileName,"RECREATE");

  //Number of training sets in the tree
  int nSets = inTree->GetEntries();

  //Loop over all of the training sets
  for(Int_t i=0;i<nSets;i++){
    if (i%1000 == 0) cout <<"Event = "<<i<<endl; //Print out every 1000 events
    inTree->GetEntry(i);

    //Set the input layer data from the tree
    layerVec[0].nodeVec[0].a = myInputTree.x1;
    layerVec[0].nodeVec[1].a = myInputTree.x2;

    //Propogate the data forward through the layers and nodes within each layer
    for (int layerIndex = 1; layerIndex < nLayers; layerIndex++) {
      for (int nodeIndex = 0; nodeIndex < netDim[layerIndex]; nodeIndex++) {
	int nW = layerVec[layerIndex].nodeVec[nodeIndex].wVec.size();
	double zVal = 0;
	for (int wIndex = 0; wIndex < nW; wIndex++) {//Create w.x dot product within node
	  double wVal = layerVec[layerIndex].nodeVec[nodeIndex].wVec[wIndex];
	  //NOTE xVal = activation in previous layer (l-1)
	  double xVal = *layerVec[layerIndex].nodeVec[nodeIndex].xVec[wIndex];
	  zVal += wVal*xVal;
	}
	zVal += layerVec[layerIndex].nodeVec[nodeIndex].b; //z = w.x + b
	//Load z and a into the node data 
	layerVec[layerIndex].nodeVec[nodeIndex].z = zVal;
	layerVec[layerIndex].nodeVec[nodeIndex].a = sigmoid(zVal);//a = sigmoid(z)
      }
    }


    //Load the truth values at output layer
    layerVec[outputLayerIndex].nodeVec[0].y = myInputTree.tVal1;
    layerVec[outputLayerIndex].nodeVec[1].y = myInputTree.tVal2;

    //find delta at output layer
    int nNode = layerVec[outputLayerIndex].nodeVec.size();
    for (int nodeIndex = 0; nodeIndex < nNode; nodeIndex++){ 
      double aVal = layerVec[outputLayerIndex].nodeVec[nodeIndex].a;
      double yVal = layerVec[outputLayerIndex].nodeVec[nodeIndex].y;
      double zVal = layerVec[outputLayerIndex].nodeVec[nodeIndex].z;
      double deltaVal = (aVal - yVal)*sigmoidPrime(zVal);
      layerVec[outputLayerIndex].nodeVec[nodeIndex].delta = deltaVal;
    }

    //Back propogate delta from output layer down to first layer above the input layer
    for (int layerIndex = outputLayerIndex - 1; layerIndex > 0; layerIndex--) {//layer l
      for (int nodeIndexJ = 0; nodeIndexJ < netDim[layerIndex]; nodeIndexJ++) {//node j
	int layerIndexP1 = layerIndex+1;//Layer above. Note P1 = plus one
	double delta_j_l = 0;
	for (int nodeIndexK = 0; nodeIndexK < netDim[layerIndexP1]; nodeIndexK++) {//node k in P1
	  int wIndexJ         = nodeIndexJ;
	  double w_k_j_lp1    = layerVec[layerIndexP1].nodeVec[nodeIndexK].wVec[wIndexJ];
	  double delta_k_lp1  = layerVec[layerIndexP1].nodeVec[nodeIndexK].delta;
	  double z_j_l        = layerVec[layerIndex].nodeVec[nodeIndexJ].z;
	  delta_j_l          += w_k_j_lp1*delta_k_lp1*sigmoidPrime(z_j_l);
	}
	layerVec[layerIndex].nodeVec[nodeIndexJ].delta = delta_j_l;
      }
    }

    //Gradient descent
    double etaVal = 1; //Learning rate
    double mVal   = 1.0;
    for (int layerIndex = outputLayerIndex; layerIndex > 0; layerIndex--) {//layer l
      for (int nodeIndexJ = 0; nodeIndexJ < netDim[layerIndex]; nodeIndexJ++) {//node j
	int nW = layerVec[layerIndex].nodeVec[nodeIndexJ].wVec.size();
	//k index represent nodes from l-1 layer and the weights have indices for current 
	//layer (l), current node (j) and previous layer (k)
	for (int wIndexK   = 0; wIndexK < nW; wIndexK++) {
	  double w_j_k_l   = layerVec[layerIndex].nodeVec[nodeIndexJ].wVec[wIndexK];
	  double delta_j_l = layerVec[layerIndex].nodeVec[nodeIndexJ].delta;
	  double b_j_l     = layerVec[layerIndex].nodeVec[nodeIndexJ].b;
	  //NOTE xVal = pointer to a_k_l-1
	  double xVal   = *layerVec[layerIndex].nodeVec[nodeIndexJ].xVec[wIndexK];
	  //RESET Weights and threshold
	  w_j_k_l = w_j_k_l - (etaVal/mVal)*delta_j_l*xVal;
	  b_j_l   = b_j_l   - (etaVal/mVal)*delta_j_l;
	  layerVec[layerIndex].nodeVec[nodeIndexJ].wVec[wIndexK] = w_j_k_l;
	  layerVec[layerIndex].nodeVec[nodeIndexJ].b = b_j_l;
	}
      }
    }
  }//FINISHED TRAINING


  cout<<"Node 0: y = ("<<-layerVec[1].nodeVec[0].wVec[0]/layerVec[1].nodeVec[0].wVec[1]<<")x + "<<
    -layerVec[1].nodeVec[0].b/layerVec[1].nodeVec[0].wVec[1]<<endl;

  cout<<"Node 1: y = ("<<-layerVec[1].nodeVec[1].wVec[0]/layerVec[1].nodeVec[1].wVec[1]<<")x + "<<
    -layerVec[1].nodeVec[1].b/layerVec[1].nodeVec[1].wVec[1]<<endl;


  //TEST->
  for (int i = 0; i<10000; i++) {
    double x1Val = myRandGen->Uniform(0.0,1.0);
    double x2Val = myRandGen->Uniform(0.0,1.0);
    
    //Set the input layer data
    layerVec[0].nodeVec[0].a = x1Val;
    layerVec[0].nodeVec[1].a = x2Val;
    for (int layerIndex = 1; layerIndex < nLayers; layerIndex++) {
      for (int nodeIndex = 0; nodeIndex < netDim[layerIndex]; nodeIndex++) {
	int nW = layerVec[layerIndex].nodeVec[nodeIndex].wVec.size();
	double zVal = 0;
	for (int wIndex = 0; wIndex < nW; wIndex++) {//Create w.x dot product within node
	  double wVal = layerVec[layerIndex].nodeVec[nodeIndex].wVec[wIndex];
	  //NOTE xVal = pointer to a_k_l-1
	  double xVal = *layerVec[layerIndex].nodeVec[nodeIndex].xVec[wIndex];
	  zVal += wVal*xVal;
	}
	zVal += layerVec[layerIndex].nodeVec[nodeIndex].b; //z = w.x + b
	layerVec[layerIndex].nodeVec[nodeIndex].z = zVal;
	layerVec[layerIndex].nodeVec[nodeIndex].a = sigmoid(zVal);//a = sigmoid(z)
      }
    }
    int y1 = 0;
    int y2 = 0;
    double a1 = layerVec[outputLayerIndex].nodeVec[0].a;
    double a2 = layerVec[outputLayerIndex].nodeVec[1].a;
    if (a1 > 0.5) y1 = 1; 
    if (a2 > 0.5) y2 = 1; 
    if (y1 == 0 && y2 == 0) hOut0->Fill(x1Val,x2Val); 
    if (y1 == 0 && y2 == 1) hOut1->Fill(x1Val,x2Val); 
    if (y1 == 1 && y2 == 0) hOut2->Fill(x1Val,x2Val); 
    if (y1 == 1 && y2 == 1) hOut3->Fill(x1Val,x2Val); 
  }
  
  outFile.cd();
  hOut0->Write();
  hOut1->Write();
  hOut2->Write();
  hOut3->Write();
  outFile.Write();
  return 0;
}

double sigmoid(double z){
  double outVal = 1/(1+exp(-z));
  return outVal;
}
double sigmoidPrime(double z){
  double outVal = sigmoid(z)*(1 - sigmoid(z));
  return outVal;
}

void printUsage(){
  fprintf(stderr,"\nUsage:");
  fprintf(stderr,"\nmakeTree [-switches]\n");
  fprintf(stderr,"\nSWITCHES:\n");
  fprintf(stderr,"-h\tPrint this message\n");
  fprintf(stderr,"-o<arg>\tOutFileName. Default is tpolOutFile.root\n");
  
  std::cout<<"The current default operation is equivalent to the command:"<<std::endl;
  std::cout<<"makeTree -ooutFile.root"<<std::endl;
  std::cout<<"where \"inputFile\" must be suplied by the user\n"<<std::endl;
  
  exit(0);
}



